import { __name } from './chunk-O6YSETKJ.mjs';
import { Cursor } from './cursor';

async function cursorPaginationFind(em, entityName, pagination, where = {}, options = {}) {
  if (pagination.after != null && pagination.before != null) {
    throw new Error(`Cannot use both 'after' and 'before' at the same time`);
  }
  if (pagination.first != null && pagination.last != null) {
    throw new Error(`Cannot use both 'first' and 'last' at the same time`);
  }
  if (pagination.first != null && pagination.first < 0) {
    throw new Error(`'first' must be greater than or equal to 0`);
  }
  if (pagination.last != null && pagination.last < 0) {
    throw new Error(`'last' must be greater than or equal to 0`);
  }
  if (pagination.orderBy != null) {
    const emptyArray = Array.isArray(pagination.orderBy) && pagination.orderBy.length === 0;
    const emptyObject = !Array.isArray(pagination.orderBy) && Object.keys(pagination.orderBy).length === 0;
    if (emptyArray || emptyObject) {
      throw new Error(`'orderBy' must be a non-empty object or array`);
    }
  }
  const forwardPagination = pagination.after != null && pagination.first != null;
  if (forwardPagination) {
    const after = pagination.after;
    const first = pagination.first;
    const cursor = Cursor.fromBase64(after);
    const paginationWhere = {
      $and: [
        {
          ...where
        },
        {
          $or: cursor.buildWhereOr({
            direction: "forward"
          })
        }
      ]
    };
    const paginationOptions = {
      ...options,
      limit: first + 1,
      orderBy: cursor.buildOrderBy({
        direction: "forward"
      })
    };
    const [totalCount, entities] = await Promise.all([
      em.count(entityName, where),
      em.find(entityName, paginationWhere, paginationOptions)
    ]);
    let hasNextPage = false;
    if (entities.length === first + 1) {
      hasNextPage = true;
      entities.pop();
    }
    const orderBy = cursor.buildOrderBy({
      direction: "forward"
    });
    const edges = entities.map((entity) => {
      return {
        cursor: Cursor.fromEntity(entity, orderBy).toBase64(),
        node: entity
      };
    });
    return {
      totalCount,
      edges,
      pageInfo: {
        hasNextPage,
        hasPreviousPage: true,
        startCursor: edges[0]?.cursor,
        endCursor: edges[edges.length - 1]?.cursor
      }
    };
  }
  const backwardPagination = pagination.before != null && pagination.last != null;
  if (backwardPagination) {
    const before = pagination.before;
    const last = pagination.last;
    const cursor = Cursor.fromBase64(before);
    const paginationWhere = {
      $and: [
        {
          ...where
        },
        {
          $or: cursor.buildWhereOr({
            direction: "backward"
          })
        }
      ]
    };
    const paginationOptions = {
      ...options,
      limit: last + 1,
      orderBy: cursor.buildOrderBy({
        direction: "backward"
      })
    };
    const [totalCount, entities] = await Promise.all([
      em.count(entityName, where),
      em.find(entityName, paginationWhere, paginationOptions)
    ]);
    let hasPreviousPage = false;
    if (entities.length === last + 1) {
      hasPreviousPage = true;
      entities.pop();
    }
    const orderBy = cursor.buildOrderBy({
      direction: "forward"
    });
    const edges = entities.reverse().map((entity) => {
      return {
        cursor: Cursor.fromEntity(entity, orderBy).toBase64(),
        node: entity
      };
    });
    return {
      totalCount,
      edges,
      pageInfo: {
        hasNextPage: true,
        hasPreviousPage,
        startCursor: edges[0]?.cursor,
        endCursor: edges[edges.length - 1]?.cursor
      }
    };
  }
  const noCursorPagination = pagination.orderBy != null && pagination.first != null;
  if (noCursorPagination) {
    const orderBy = pagination.orderBy;
    const first = pagination.first;
    const [totalCount, entities] = await Promise.all([
      em.count(entityName, {
        ...where
      }),
      em.find(entityName, {
        ...where
      }, {
        ...options,
        limit: pagination.first,
        orderBy: pagination.orderBy
      })
    ]);
    const edges = entities.map((entity) => {
      return {
        cursor: Cursor.fromEntity(entity, orderBy).toBase64(),
        node: entity
      };
    });
    return {
      totalCount,
      edges,
      pageInfo: {
        hasNextPage: totalCount > first,
        hasPreviousPage: false,
        startCursor: edges[0]?.cursor,
        endCursor: edges[edges.length - 1]?.cursor
      }
    };
  }
  throw new Error("Invalid pagination parameters");
}
__name(cursorPaginationFind, "cursorPaginationFind");

export { cursorPaginationFind };
